Program CZtoCZ;                    {Copyright Hans GINZEL, Praha 20.6.1996}

uses WinDos, WinCrt, Strings;      {You can compile it into Dos & Windos Target}

const
  Version               = '1.14w'; {Can't read CodePage}{-L, -U}{In = Out}{Last update: 23.7.1996}
  LongintArrayLength    = 32;      {< Word}
  HalfStringLength      = 128;     {Don't change this!}
  StringLength          = 255;
  StartOfLowerCaseChars = 48;      {in CodeStrings}
  LinesPerPage          = 30;
  NL                    = Chr(13)+Chr(10);
  TempFileName          = '~CZtoCZ~.tmp'; {if <InFile> = <OutFile> then rename <InFile> TempFileName}

type
  TLongintArray  = array [0..LongintArrayLength] of Longint;
  THalfString = String[HalfStringLength];

var
  InCodePage, OutCodePage: array [0..StringLength] of Char;
  InFileName, OutFileName: String;
  InCodePageStr, OutCodePageStr: THalfString;
  InFile, OutFile: Text;
  CountOfLines:    Longint;
  ErrorLines:      TLongintArray;
  ErrorCharacters: String;
  E, L, U:         Boolean; {echo, lowercase, uppercase}
  i, j:            Integer;

procedure WriteUsage;
  begin
    WriteLn('CZtoCZ [@] <From> <To> [<Options>] [[<Options>]...] <InFile> <OutFile>');
    WriteLn('  <From>    ... the coding page of the  input file ([Asc|Iso|Kam|Lat|Win|Usr])');
    WriteLn('  <To>      ... the coding page of the output file ([Asc|Iso|Kam|Lat|Win|Usr])');
    WriteLn('  <InFile>  ... the name of the  input file; if empty,  StdInput will be used.');
    WriteLn('  <OutFile> ... the name of the output file; if empty, StdOutput will be used.');
    WriteLn('  <Options> :   -l ... UpperCase -> LowerCase,  -u ... LowerCase -> UpperCase,');
    WriteLn('    @       ... echo off.                      Instead of ''-'' you can use ''/''.');
    WriteLn('Converts <InFile> into <OutFile> using permutation of characters given by');
    WriteLn('  files <From> & <To> and by options.');
    {<from> & <To> are names of the files containing the coding string}
  end;

procedure WriteErrors(var Lines: TLongintArray; var Characters: String);
  var i: Word;
  begin
    if Lines[0] = 0 then Exit;
    WriteLn('The input file contains ', Lines[0], ' bad characters. Bad character is character');
    WriteLn('  with code greather then 127 that is not in input coding page string (file).');
    if Length(Characters) <> StringLength then Write('Here are: ')
    else Write('Some of them are here: ');
    Write ('''', Characters[1], '''');
    i := 2;
    while not(i > Length(Characters)) do begin
      Write (', ''', Characters[i], ''''); Inc(i);
    end; WriteLn(';');
    if Lines[LongintArrayLength] <> 0 then Write('you can find them for example in these lines: ')
    else Write('you can find them in these lines: ');
    Write (Lines[1]);
    i := 2;
    while not(i > LongintArrayLength) and (Lines[i] <> 0) do begin
      Write (', ', Lines[i]); Inc(i);
    end; WriteLn('.');
  end;

procedure ReadCodePage(CodePage: PChar; var CodePageStr: THalfString; var ErrorStr: String);
  var
    CodePageName, PathWhereIam: array [0..StringLength] of Char;
    CodePageFile: Text;
    i: Word; C: Char;
  begin
    FileSplit(GetArgStr(PathWhereIam, 0, StringLength), PathWhereIam, nil, nil);
    FileSearch(CodePageName, CodePage, StrCat(StrCat(PathWhereIam, ';'), GetEnvVar('PATH')));
    if not(CodePageName[0] <> #0) then
      ErrorStr := 'Error! Sorry, but I can not find the file ''' + StrPas(CodePage) + '''' +
           + NL + '         containing the coding string matching coding page of an in/out file.';
    if ErrorStr <> '' then Exit;
    Assign(CodePageFile, CodePageName);
    {$I-}  Reset(CodePageFile);  {$I+}
    if IOResult <> 0 then ErrorStr := 'Error! Sorry, but I can not open the file ''' + StrPas(CodePage) + '''' +
           + NL + '         containing the coding string matching coding page of an in/out file.';
    if ErrorCharacters <> '' then Exit;
    CodePageStr := ''; i := 0;
    while not EoF(CodePageFile) and (i < HalfStringLength) do begin
      Read(CodePageFile, C); CodePageStr := CodePageStr + C; Inc(i);
    end;
    if not EoF(CodePageFile) then begin
      Str(HalfStringLength, PathWhereIam);
      ErrorStr := 'Error! Sorry, but the file ''' + StrPas(CodePageName) + ''' is too long.' + NL +
        + 'I expect a file containing the coding string with length maximal ' + StrPas(PathWhereIam) + ' characters.'
    end;
    {if ErrorStr <> '' then Exit;}
    Close(CodePageFile);
  end;

procedure Replace(var InStr, OutStr: THalfString; var InFile, OutFile: Text; Echo, L, U: Boolean;
            {out} var N: Longint; var Lines: TLongintArray; var Characters: String);
  {In Characters will be stored some Bad Characters or the ErrorString.
   Bad Character is character with code greather then 127 and not in InCodeStr.
   In Lines will be stored first LongintArrayLength line numbers containing Bad Characters or zeros.
   Lines[0] := coutnt of Bad Characters. N is count of all readed lines.}
var
  C: Char;  i: Word; Permutation: String;
  InStrLength, OutStrLength: Byte;
begin
  for i := 0 to LongintArrayLength do Lines[i] := 0;
  Characters := '';
  N := 0;  {line counter}
  InStrLength  := Length(InStr);
  OutStrLength := Length(OutStr);
  if InStrLength < OutStrLength then
    Characters := 'Error! Length of the coding string matching coding page of the input file is' +
           + NL + '         less then length of the coding string of the output file.';
  if InStrLength > OutStrLength then
    Characters := 'Error! Length of the coding string matching coding page of the input file is' +
           + NL + '         greather then length of the coding string of the output file.';
  if Characters <> '' then Exit;
  if (L or U) and (InStrLength < 2*StartOfLowerCaseChars) then begin
    Str(StartOfLowerCaseChars, Permutation);
    Characters := 'Error! Sorry, but I can not convert the file to upper or lowercase characters.' +
           + NL + '  I assumme that the uppercase characters are in the In & OutStr at positions' +
           + NL + '  1..' + Permutation + ' and the matching lowercase characters positions ' +
                + Permutation + '+1..2*' + Permutation + ' .';
  end;
  if Characters <> '' then Exit;
  for i :=   1 to 127 do Permutation[i] := Chr(i); { ------ Permutation ------ }
  for i := 128 to 255 do Permutation[i] := Chr(0);
  for i := 1 to InStrLength  do Permutation[Ord(InStr[i])] := OutStr[i];
  if L then for i := 1 to StartOfLowerCaseChars do                     {LowerCase}
                                Permutation[Ord(InStr[i])] := OutStr[i+StartOfLowerCaseChars];
  if U then for i := 1 to StartOfLowerCaseChars do                     {UpperCase}
            Permutation[Ord(InStr[i+StartOfLowerCaseChars])] := OutStr[i];
  N := 1; { ------------------------------------------------------------------ }
  while not EoF(InFile) do begin { ---------------- Main loop  --------------- }
      Read(InFile, C);             {ctu to po znaku, doufam, ze cache to spravi}
      if Permutation[Ord(C)] <> #0 then Write(OutFile, Permutation[Ord(C)])
      else begin                                     {adding error informations}
          Write(OutFile, C);
          i := 1;
          while (i < Length(Characters)) and (Characters[i] <> C) do Inc(i);
          if C <> Characters[i] then Characters := Characters + C;
          i := 1;
          while (i < LongintArrayLength) and (Lines[i] <> N) and (Lines[i] <> 0) do Inc(i);
          if N <> Lines[i] then if i < LongintArrayLength then Lines[i] := N;
          Inc(Lines[0]);
        end;
      if C = Chr(10) then begin                          {increase line counter}
        Inc(N); if Echo then if N mod LinesPerPage = 0 then Write(' [', N div LinesPerPage, ']');
      end;
  end; { --------------------------------------------------------------------- }
  if Echo then if N mod LinesPerPage <> 0 then Write(' [', N div LinesPerPage + 1, ']');
end;

begin
  { ================================== Help ================================== }
  ErrorCharacters := '';
  if (ParamStr(1) = '/?') or (ParamStr(1) = '-?') or (ParamStr(1) = '?') or
     (ParamStr(1) = '/h') or (ParamStr(1) = '-h') or (ParamStr(1) = 'h')
  then begin  WriteUsage; Exit  end;
  { ========================= Commandline parameters ========================= }
  i := 0;
  Inc(i); GetArgStr(InCodePage, i, StringLength);
  E := InCodePage[1] = '@';
  if E then begin  Inc(i); GetArgStr(InCodePage, i, StringLength);  end;
  if not E then WriteLn('This is Convertor CZtoCZ, Version ', Version); {Message}
  if not(InCodePage[0] <> #0) then
    ErrorCharacters := 'Error! Sorry, but I expect the coding page of the input file.' + NL + 'Use parametr /? for help.';
  if ErrorCharacters <> '' then begin  WriteLn(ErrorCharacters); Exit  end;
  Inc(i); GetArgStr(OutCodePage, i, StringLength);
  if not(OutCodePage[0] <> #0) then
    ErrorCharacters := 'Error! Sorry, but I expect the coding page of the output file.' + NL + 'Use parametr /? for help.';
  if ErrorCharacters <> '' then begin  WriteLn(ErrorCharacters); Exit  end;
  L := False; U := False;
  Inc(i); InFileName  := ParamStr(i);
  while (InFileName[1] = '-') or (InFileName[1] = '/') do begin
    for j := 2 to Length(InFileName) do begin
      L := L or (InFileName[j] = 'l') or (InFileName[j] = 'L');
      U := U or (InFileName[j] = 'u') or (InFileName[j] = 'U');
      if (InFileName[j] <> 'l') and (InFileName[j] <> 'L') and
         (InFileName[j] <> 'u') and (InFileName[j] <> 'U') then
        ErrorCharacters := 'Error! Sorry, I don''t know this option: ''' + InFileName[j] + '''. Use parametr /? for help.';
      if ErrorCharacters <> '' then WriteLn(ErrorCharacters);
      ErrorCharacters := '';
    end;
  Inc(i); InFileName  := ParamStr(i);
  end;
  Inc(i); OutFileName := ParamStr(i);
  { ============================ Opening the files =========================== }
  ReadCodePage(InCodePage, InCodePageStr, ErrorCharacters);
  if ErrorCharacters <> '' then begin  WriteLn(ErrorCharacters); Exit  end;
  ReadCodePage(OutCodePage, OutCodePageStr, ErrorCharacters);
  if ErrorCharacters <> '' then begin  WriteLn(ErrorCharacters); Exit  end;
  Assign(InFile, InFileName);
  {$I-}  Reset(InFile);  {$I+}
  if IOResult <> 0 then ErrorCharacters := 'Error! Sorry, but I can not open the input file ''' + InFileName + '''';
  if ErrorCharacters <> '' then begin  WriteLn(ErrorCharacters); Exit  end;
  if (InFileName = OutFileName) and (InFileName <> '') then begin
    {$I-} Assign (OutFile, TempFileName); Erase(OutFile); {$I+} j := IOResult;
    {$I-} Close(InFile); Rename(InFile, TempFileName); Reset(InFile);  {$I+}
    if IOResult <> 0 then ErrorCharacters := 'Error! Sorry, but I can not rename the input file.' +
                                      + NL + '         Use a different name for the output file.';
    if ErrorCharacters <> '' then begin  WriteLn(ErrorCharacters); {$I-}  Close(InFile); {$I+} Exit  end;
  end;
  Assign(OutFile, OutFileName);
  {$I-}  Rewrite(OutFile);  {$I+}
  if IOResult <> 0 then ErrorCharacters := 'Error! Sorry, but I can not open the output file ''' + OutFileName + '''';
  if ErrorCharacters <> '' then begin  WriteLn(ErrorCharacters); {$I-}  Close(InFile); {$I+} Exit  end;
  { ================================= Running ================================ }
  if not E then Write('Input (coding page ''', InCodePage, ''') (', InFileName);
  CountOfLines := 0;
  Replace(InCodePageStr, OutCodePageStr, InFile, OutFile, not E, L, U, CountOfLines, ErrorLines, ErrorCharacters);
  if not E then WriteLn(')');
  Close(OutFile);
  Close(InFile);
  if not E then WriteLn('Output written on ', OutFileName, ' (coding page ''', OutCodePage, '''; ', CountOfLines, ' lines).');
  if ErrorLines[0] <> 0 then WriteErrors(ErrorLines, ErrorCharacters) else
    if ErrorCharacters <> '' then WriteLn(ErrorCharacters);
end.
